'''
gtrace.beam

A module to define GaussianBeam class.

'''

#{{{ Import modules

import numpy as np
pi = np.pi
array = np.array
sqrt = np.lib.scimath.sqrt
from numpy.linalg import norm

from traits.api import HasTraits, Int, Float, CFloat, CComplex, CArray, List, Str

from .unit import *
import gtrace.optics as optics
import gtrace.optics.geometric
from gtrace.optics.gaussian import q2zr, q2w, q2R, optimalMatching
import copy
#import gtrace.sdxf as sdxf
import gtrace.draw as draw
import scipy.optimize as sopt

#}}}

#{{{ Author and License Infomation

#Copyright (c) 2011-2012, Yoichi Aso
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

__author__ = "Yoichi Aso"
__copyright__ = "Copyright 2011-2021, Yoichi Aso"
__credits__ = ["Yoichi Aso"]
__license__ = "BSD"
__version__ = "0.2.4"
__maintainer__ = "Yoichi Aso"
__email__ = "yoichi.aso@nao.ac.jp"
__status__ = "Beta"

#}}}

#{{{ GaussianBeam Class

class GaussianBeam(HasTraits):
    '''
    This is a class to represent a Gaussian beam.
    A GaussianBeam object has its origin (pos) and a propagation
    direction (dirVect or dirAngle).
    A GaussianBeam is characterized by q-parameter(s) at its origin.
    The beam can be either circular or elliptic. In order to deal with
    elliptic beams, some parameters are stored in pairs like (q0x, q0y).
    x and y denote the axes of the cross section of the beam. x-axis is
    parallel to the paper and the y-axis is perpendicular to the paper.

    A beam object can be propagated through a free space or made
    to interact with an optics.

    As a beam propagate through optical system, optical distance and
    Gouy phase are accumerated.

    Attributes
    ----------
    q : complex
        q-parameter of the beam. If the beam is eliptic, q is the q-parameter
        of the best matching circular mode.
    qx : complex
        q-parameter of the beam in the x-direction.
    qy : complex
        q-parameter of the beam in the y-direction.
    pos : array
        Position of the beam origin (x, y).
    dirVect : array
        Propagation direction vector.
    dirAngle : float
        Propagation direction angle measured from the positive x-axis.
    length : float
        Length of the beam (used for DXF export)
    layer : str
        Layer name of the beam when exported to a DXF file.
    name : str
        Name of the beam
    wl : float
        Wavelength in vacuum. Not the wavelength in the medium.
    n : float
        Index of refraction of the medium the beam is passing through.
    P : float
        Power.
    wx : float
        Beamwidth in x-direction.
    wy : float
        Beamwidth in y-direction.
    optDist : float
        Accumulated optical distance.
    Gouyx : float
        Accumulated Gouy phase in x-direction.
    Gouyy : float
        Accumulated Gouy phase in y-direction.
    Mx : array
        ABCD matrix in x-direction.
        This is a 2x2 matrix representing the product
        of ABCD transformations applied to this beam.
        It defaults to an identity matrix.
        Whenever a beam experience an ABCD matrix
        transformation, such as propagation in the space
        or reflection by a curved mirror, the applied ABCD
        matrix is multiplied to this matrix, so that we can
        keep track of what kind of transformations were
        made during beam propagation.
    My : array
        ABCD matrix in y-direction. The meaning is the same as Mx.
    departSurfAngle : None
        The angle formed by x-axis and the normal vector of
        the surface from which the beam is departing.
        Default is None. Used by the drawing routine.
    departSurfInvROC : None
        Inverse of the ROC of the surface from which the beam is departing.
        The ROC is positive for a concave surface seen from
        the beam side.
        Default is None. Used by the drawing routine.
    incSurfAngle : None
        The angle formed by the x-arm and the normal vector of
        the surface to which the beam is incident.
        Default is None. Used by the drawing routine.
    incSurfInvROC : None
        Inverse of the ROC of the surface to which the beam is incident.
        The ROC is positive for a concave surface seen from
        the beam side.
        Default is None. Used by the drawing routine.
    stray_order : int
        An integer indicating if this beam is a stray light or not.
        The default value is 0. Every time a beam is reflected by an AR surface
        or transmits an HR surface, this couter is increased by 1.
    '''
#{{{ Traits Definitions

    name = Str()
    wl = CFloat(1064.0*nm)  #Wavelength
    P = CFloat(1*W)  #Power
    q = CComplex()  #q-parameter at the origin (best matching circular mode)
    qx = CComplex(1j)  #q-parameter at the origin (x-direction)
    qy = CComplex(1j)  #q-parameter at the origin (y-direction)
    qrx = CComplex()  #Reduced q-parameter at the origin (x-direction)
                                  # qrx = qx/n
    qry = CComplex()  #Reduced q-parameter at the origin (x-direction)
                                  # qry = qy/n
    Gouyx = CFloat(0.0) #Accumurated Gouy phase
    Gouyy = CFloat(0.0) #Accumurated Gouy phase
    wx = CFloat()
    wy = CFloat()
    n = CFloat(1.0)

    pos = CArray(dtype='float64', shape=(2,))
    length = CFloat(1.0)
    layer = Str()
    dirVect = CArray(dtype='float64', shape=(2,))
    dirAngle = CFloat()
    optDist = CFloat(0.0)

    Mx = CArray(value=[[1,0],[0,1]],dtype='float64', shape=(2,2))
    My = CArray(value=[[1,0],[0,1]],dtype='float64', shape=(2,2))

#}}}

#{{{ __init__

    def __init__(self, q0=1j*2*pi/(1064*nm)*1e-6/2, q0x=False, q0y=False,
                 pos=[0.0,0.0], length=1.0, dirAngle=0.0, dirVect=None,
                 wl=1064*nm, P=1*W, n=1.0, name="Beam", layer='main_beam'):
        '''Constructor

        Parameters
        ----------
        q0 : complex, optional
            q-parameter. If q0x or q0y is not given, this parameter is used
            as both qx and qy.
            Defaults 1j*2*pi/(1064*nm)*1e-6/2.
        q0x : complex, optional
            q-parameter in x-direction.
            Defaults False. If False, set to q0.
        q0y : complex, optional
            q-parameter in y-direction.
            Defaults False. If False, set to q0.
        pos : array, optional
            Position of the origin of the beam (x, y).
            Defaults [0.0, 0.0].
        length : float, optional
            Length of the beam (used for DXF export).
            Defaults 1.0.
        dirAngle : float, optional
            Propagation direction angle measured from the positive x-axis.
            Defaults 0.0.
        dirVect : array, optional
            Propagation direction vector.
            Defaults None.
        wl : float, optional.
            Wavelength.
            Defaults 1064nm.
        P : float, optional.
            Power.
            Defaults 1W.
        n : float, optional
            Index of refraction of the medium the beam is passing through.
            Defaults 1.0.
        name : str, optional
            Name of the beam.
            Defaults "Beam".
        layer : str, optional
            Layer name of the beam when exported to a DXF file.
            Defaults "main_beam".
        '''
        self.wl = wl
        self.P = P
        self.pos = pos
        self.length = length
        self.name = name
        self.layer = layer
        self.n = n

        if q0x:  ## Possible bug. If q0x is True then self.qx becomes True.
            self.qx = q0x
        else:
            self.qx = q0

        if q0y:
            self.qy = q0y
        else:
            self.qy = q0

        if dirVect is not None:
            self.dirVect = dirVect
        else:
            self.dirAngle = dirAngle
            self._dirAngle_changed(0,0)

        self.optDist = 0.0

        self.departSurfAngle = None
        self.departSurfInvROC = None
        self.incSurfAngle = None
        self.incSurfInvROC = None
        self.stray_order = 0

#}}}

#{{{ copy

    def copy(self):
        '''
        Make a deep copy.
        '''
        b = copy.deepcopy(self)
        b.qrx = self.qrx
        b.qry = self.qry
        return b

#}}}

#{{{ propagate

    def propagate(self, d):
        '''
        Propagate the beam by a distance d from the current position.
        self.n is used as the index of refraction.
        During this process, the optical distance traveled is added
        to self.optDist.
        self.Goux and self.Gouyy are also updated to record the Gouy
        phase change.

        Parameters
        ----------
        d: float
            Distance.
        '''
        qx0 = self.qx
        qy0 = self.qy

        ABCD = np.array([[1.0, d/self.n],[0.0, 1.0]])
        #ABCD = np.array([[1.0, d/self.n**2],[0.0, 1.0]])
        self.ABCDTrans(ABCD)
        self.pos = self.pos + self.dirVect*d

        #Increase the optical distance
        self.optDist = self.optDist + self.n*d

        #Increase the Gouy phase
        self.Gouyx = self.Gouyx + np.arctan(np.real(self.qx)/np.imag(self.qx))\
                     - np.arctan(np.real(qx0)/np.imag(qx0))

        self.Gouyy = self.Gouyy + np.arctan(np.real(self.qy)/np.imag(self.qy))\
                     - np.arctan(np.real(qy0)/np.imag(qy0))

#}}}

#{{{ ABCD Trans
    def ABCDTrans(self, ABCDx, ABCDy=None):
        '''
        Apply ABCD transformation to the beam.

        Parameters
        ----------
        ABCDx : array
            ABCD matrix for x-direction.
        ABCDy : array or None, optional.
            ABCD matrix for y-direction.
            Defaults None. If None, set to ABCDx.
        '''
        if ABCDy is None:
            ABCDy = ABCDx

        #Update q-parameters
        self.qrx = (ABCDx[0,0]*self.qrx + ABCDx[0,1])/(ABCDx[1,0]*self.qrx + ABCDx[1,1])
        self.qry = (ABCDy[0,0]*self.qry + ABCDy[0,1])/(ABCDy[1,0]*self.qry + ABCDy[1,1])

        #Update Mx and My
        self.Mx = np.dot(ABCDx, self.Mx)
        self.My = np.dot(ABCDy, self.My)

#}}}

#{{{ rotate

    def rotate(self, angle, center=False):
        '''
        Rotate the beam around 'center'.
        If center is not given, the beam is rotated
        around self.pos.

        Parameters
        ----------
        angle : float
            Rotation angle in radians.
        center : array or boolean.
            Center of rotation. Should be an array of shape(2,).
            Defaults False.
        '''
        if center is not False:
            center = np.array(center)
            pointer = self.pos - center
            pointer = optics.geometric.vector_rotation_2D(pointer, angle)
            self.pos = center + pointer

        self.dirAngle = self.dirAngle + angle

#}}}

#{{{ Translate

    def translate(self, trVect):
        '''
        Translate the beam by the direction and the distance
        specified by a vector.

        Parameters
        ----------
        trVect : array
            A vector to specify the translation direction and
            distance. Should be an array of shape(2,)
        '''
        trVect = np.array(trVect)
        self.pos = self.pos + trVect

#}}}

#{{{ Flip
    def flip(self, flipDirVect=True):
        '''
        Change the propagation direction of the beam
        by 180 degrees.
        This is equivalent to the reflection of the beam
        by a spherical mirror with the same ROC as the beam.

        If optional argument flipDirVect is set to False,
        the propagation direction of the beam is not changed.

        Parameters
        ----------
        flipDirVect : boolean, optional
            Flip propagation direction.
            Defaults True.
        '''
        self.qx = -np.real(self.qx)+1j*np.imag(self.qx)
        self.qy = -np.real(self.qy)+1j*np.imag(self.qy)
        if flipDirVect:
            self.dirVect = - self.dirVect

#}}}

#{{{ width

    def width(self, dist):
        '''
        Returns the beam width at a distance dist
        from the origin of the beam.
        The width is the radius where the light power becomes 1/e^2.

        Parameters
        ----------
        dist : float
            Distance.

        Returns
        -------
        (float, float)
            The width of the beam in x and y direction.
        '''
        dist = np.array(dist)
        k = 2*pi/(self.wl/self.n)
        qx = self.qx + dist
        qy = self.qy + dist

        return (np.sqrt(-2.0/(k*np.imag(1.0/qx))), np.sqrt(-2.0/(k*np.imag(1.0/qy))))

#}}}

#{{{ R

    def R(self, dist=0.0):
        '''
        Returns the beam ROC at a distance dist
        from the origin of the beam.

        Parameters
        ----------
        dist : float, optional
            Distance.

        Returns
        -------
        (float, float)
            Beam ROC.
        '''
        dist = np.array(dist)
        k = 2*pi/self.wl
        qx = self.qx + dist/self.n
        qy = self.qy + dist/self.n

        return (q2R(qx), q2R(qy))

#}}}

#{{{ Waist

    def waist(self):
        '''
        Return the tuples of waist size and distance

        Returns
        -------
        dict
            {"Waist Size": (float, float), "Waist Position": (float, float)}
        '''

        #Waist positions
        dx = -np.real(self.qx)
        dy = -np.real(self.qy)

        #Waist sizes
        wx = self.width(dx)[0]
        wy = self.width(dy)[1]

        return {"Waist Size":(wx, wy), "Waist Position":(dx, dy)}

#}}}

#{{{ draw

#{{{ Main Function

    def draw(self, cv, sigma=3., mode='x', drawWidth=True,
             fontSize=False, drawPower=False, drawROC=False, drawGouy=False,
             drawOptDist=False, drawName=False, debug=False):
        '''
        Draw the beam into a DXF object.

        Parameters
        ----------
        cv : gtrace.draw.draw.Canvas
            gtrace canvas.
        sigma : float, optional
            The width of the beam drawn is sigma * (1/e^2 radius of the beam).
            The default is sigma = 3. sigma = 2.7 gives 1ppm diffraction loss.
            Defaults 3.
        mode : str, optional
            'avg', 'x', or 'y'. A beam can have different widths for x- and y-
            directions. If 'avg' is specified, the average of them are drawn.
            'x' and 'y' specifies to show the width of the respective directions.
            Defaults 'x'.
        fontSize : float, optional
            Size of the font used to show supplemental informations.
            Defaults False.
        drawWidth : boolean, optional
            Whether to draw width or not.
            Defaults True.
        drawPower : boolean, optional
            Whether to show the beam power.
            Defaults False.
        drawROC : boolean, optional
            Whether to show the ROC or not.
            Defaults False.
        drawGouy : boolean, optional
            Whether to show the Gouy phase or not.
            Defaults False.
        drawOptDist : boolean, optional
            Whether to show the accumulated optical distance or not.
            Defaults False.
        drawName : boolean, optional
            Whether draw the name of the beam or not.
            Defaults False.
        debug : boolean, optional
            Debug.
        '''
        if not fontSize:
            fontSize = self.width(self.length/2)[0]*sigma/5

        start = tuple(self.pos)
        stop = tuple(self.pos + self.dirVect * self.length)

        #Location to put texts
        #mid = self.pos + self.dirVect * self.length/2
        #side = mid+fontSize*1.2*k
        k = np.array((-self.dirVect[1], self.dirVect[0]))
        text_location = tuple(self.pos + (self.length - 10*fontSize)*self.dirVect + k*fontSize*1)


        #Draw the center line
        cv.add_shape(draw.Line(start, stop), self.layer)

        if drawWidth:
            self.drawWidth(cv, sigma, mode)

        annotation = ''
        if drawName:
            annotation = annotation+'%s '%self.name

        if drawPower:
            annotation = annotation+'P=%.2E '%self.P

        if drawROC:
            annotation = annotation+'ROCx=%.2E '%q2R(self.qx)
            annotation = annotation+'ROCy=%.2E '%q2R(self.qy)

        if drawGouy:
            annotation = annotation+'Gouyx=%.2E '%self.Gouyx
            annotation = annotation+'Gouyy=%.2E '%self.Gouyy

        if drawOptDist:
            annotation = annotation+'Optical distance=%.2E '%self.optDist

        cv.add_shape(draw.Text(text=annotation, point=text_location,
                             height=fontSize), layername='text')

        #Indicate the beam direction
        # text_location = tuple(self.pos + 10*fontSize*self.dirVect +k*fontSize*1)
        # dxf.append(sdxf.Text(text=self.name+':origin P=%.2E'%self.P, point=text_location,
        #                      height=fontSize, layer='text'))

        # text_location = tuple(self.pos + (self.length - 10*fontSize)*self.dirVect + k*fontSize*1)
        # dxf.append(sdxf.Text(text=self.name+':end P=%.2E'%self.P, point=text_location,
        #                      height=fontSize, layer='text'))

#}}}

#{{{ drawWidth

    def drawWidth(self, cv, sigma, mode):
        """Draw width on canvas.

        Parameters
        ----------
        cv : gtrace.draw.draw.Canvas
            The canvas.
        sigma : float
            The width of the beam drawn is sigma * (1/e^2 radius of the beam).
            The default is sigma = 3. sigma = 2.7 gives 1ppm diffraction loss.
        mode : str
            'avg', 'x', or 'y'. A beam can have different widths for x- and y-
            directions. If 'avg' is specified, the average of them are drawn.
            'x' and 'y' specifies to show the width of the respective directions.
        """
        #Draw the width

        #Determine the number of line segments to draw
        zr = q2zr(self.qx)
        resolution = zr/10.0
        if resolution > self.length/10.0:
            resolution = self.length/10.0

        numSegments = int(self.length/resolution)

        if numSegments > 100:
            numSegments = 100

        #q0 to be used
        if mode == 'x':
            q0 = self.qx
        elif mode == 'y':
            q0 = self.qy
        else:
            q0 = (self.qx + self.qy)/2.0

        #Determine the start z coordinates
        if self.departSurfAngle is not None:
            theta = self.departSurfAngle - self.dirAngle
            k = 2*pi/(self.wl/self.n)
            if np.abs(self.departSurfInvROC) > 1.0/100:
                R = 1.0/self.departSurfInvROC
                (z1u, z1d) = optimStartPointR(theta, R, q0, k, sigma)
            else:
                (z1u, z1d) = optimCrossPointFlat(theta, q0, k, sigma)
        else:
            z1u = 0.0
            z1d = 0.0

        #Determine the end z coordinates
        if self.incSurfAngle is not None:
            theta = np.mod(self.incSurfAngle+pi, 2*pi) - self.dirAngle
            k = 2*pi/(self.wl/self.n)
            if np.abs(self.incSurfInvROC) > 1.0/100:
                R = 1.0/self.incSurfInvROC
                (z2u, z2d) = optimEndPointR(theta, R, q0+self.length, k, sigma)
            else:
                (z2u, z2d) = optimCrossPointFlat(theta, q0+self.length, k, sigma)
        else:
            z2u = 0.0
            z2d = 0.0


        du = np.linspace(z1u, self.length+z2u, numSegments)
        dd = np.linspace(z1d, self.length+z2d, numSegments)
        au = self.width(du)
        ad = self.width(dd)
        if mode == 'x':
            wu = au[0]*sigma
            wd = ad[0]*sigma
        elif mode == 'y':
            wu = au[1]*sigma
            wd = ad[1]*sigma
        else:
            wu = sigma*(au[0]+au[1])/2
            wd = sigma*(ad[0]+ad[1])/2

        v = np.vstack((du,wu))
        v = optics.geometric.vector_rotation_2D(v, self.dirAngle)
        v = v + np.array([self.pos]).T
        cv.add_shape(draw.PolyLine(x=v[0,:], y=v[1,:]), layername=self.layer+"_width")

        v = np.vstack((dd,-wd))
        v = optics.geometric.vector_rotation_2D(v, self.dirAngle)
        v = v + np.array([self.pos]).T
        cv.add_shape(draw.PolyLine(x=v[0,:], y=v[1,:]), layername=self.layer+"_width")


#}}}

#}}}

#{{{ Notification Handlers

    def _dirAngle_changed(self, old, new):
        self.trait_set(trait_change_notify=False,
                 dirVect=array([np.cos(self.dirAngle), np.sin(self.dirAngle)]))
        self.trait_set(trait_change_notify=False,
                 dirAngle = np.mod(self.dirAngle, 2*pi))
#        self.dirVect = array([np.cos(self.dirAngle), np.sin(self.dirAngle)])
#        self.dirAngle = np.mod(self.dirAngle, 2*pi)

    def _dirVect_changed(self, old, new):
        #Normalize
        self.trait_set(trait_change_notify=False,
                 dirVect = self.dirVect/np.linalg.norm(array(self.dirVect)))
        #Update dirAngle accordingly
        self.trait_set(trait_change_notify=False,
                 dirAngle = np.mod(np.arctan2(self.dirVect[1],
                                              self.dirVect[0]), 2*pi))

    def _qx_changed(self, old, new):
        self.wx = q2w(self.qx, wl=self.wl/self.n)
        self.trait_set(trait_change_notify=False,
                 qrx = self.qx/self.n)
        self.q = optimalMatching(self.qx, self.qy)[0]

    def _qy_changed(self, old, new):
        self.wy = q2w(self.qy, wl=self.wl/self.n)
        self.trait_set(trait_change_notify=False,
                 qry = self.qy/self.n)
        self.q = optimalMatching(self.qx, self.qy)[0]

    def _qrx_changed(self, old, new):
        self.qx = self.qrx*self.n

    def _qry_changed(self, old, new):
        self.qy = self.qry*self.n

    def _n_changed(self, old, new):
        self.trait_set(trait_change_notify=False,
                 qx = self.qrx*self.n)
        self.trait_set(trait_change_notify=False,
                 qy = self.qry*self.n)

#}}}

#}}}

#{{{ Functions to determine the cross points

def optFunForStartPointR(phi, Mrot, R, q0, k, sigma, side):
    '''
    A function to return the distance between the point on
    the spherical surface at an angle phi and the beam width
    at the same z.

    Parameters
    ----------
    phi : float
        phi
    Mrot: array
        Rotational transformation.
    R : float
        R
    q0 : complex
        Beam parameter.
    k : float
        k
    sigma : float
        Beam width
    side
        side

    Returns
    -------
    float
        Distance between the point on the spherical surface at an
        angle phi and the beam width at the same z.
    '''
    cp = np.cos(phi)
    sp = np.sin(phi)
    v = R*np.array([[1-cp], [-sp]])

    a = np.dot(Mrot, v)
    z = a[0]
    w1 = a[1]*side
    w2 = sigma*np.sqrt(-2.0/(k*np.imag(1/(q0+z))))
    return (w1-w2)[0]

def optimStartPointR(theta, R, q0, k, sigma):
    '''Caltulate optimal starting point.

    Paramters
    ---------
    theta : float
        theta
    R : float
        R
    q0 : complex
        Beam parameter.
    k : float
        k
    sigma : float
        Beam width.

    Returns
    -------
    (float, float)
        Optimal starting point.
    '''
    ct = np.cos(theta)
    st = np.sin(theta)
    Mrot = np.array([[ct, -st], [st, ct]])

    phi1 = sopt.newton(optFunForStartPointR, 0, args=(Mrot, R, q0, k, sigma, 1))
    cp = np.cos(phi1)
    sp = np.sin(phi1)
    v = R*np.array([[1-cp], [-sp]])
    z1 = np.dot(Mrot, v)[0][0]

    phi2 = sopt.newton(optFunForStartPointR, 0, args=(Mrot, R, q0, k, sigma, -1))
    cp = np.cos(phi2)
    sp = np.sin(phi2)
    v = R*np.array([[1-cp], [-sp]])
    z2 = np.dot(Mrot, v)[0][0]

    return (z1,z2)

def optFunForEndPointR(phi, Mrot, R, q0, k, sigma, side):
    '''
    A function to return the distance between the point on
    the spherical surface at an angle phi and the beam width
    at the same z.

    Parameters
    ----------
    phi : float
        phi
    Mrot: array
        Rotational transformation.
    R : float
        R
    q0 : complex
        Beam parameter.
    k : float
        k
    sigma : float
        Beam width
    side
        side

    Returns
    -------
    float
        Distance between the point on the spherical surface at an
        angle phi and the beam width at the same z.
    '''
    cp = np.cos(phi)
    sp = np.sin(phi)
    v = -R*np.array([[1 - cp], [-sp]])

    a = np.dot(Mrot, v)
    z = a[0]
    w1 = a[1]*side
    w2 = sigma*np.sqrt(-2.0/(k*np.imag(1/(q0+z))))
    return (w1-w2)[0]

def optimEndPointR(theta, R, q0, k, sigma):
    '''Caltulate optimal end point.

    Paramters
    ---------
    theta : float
        theta
    R : float
        R
    q0 : complex
        Beam parameter.
    k : float
        k
    sigma : float
        Beam width.

    Returns
    -------
    (float, float)
        Optimal end point.
    '''
    ct = np.cos(theta)
    st = np.sin(theta)
    Mrot = np.array([[ct, -st], [st, ct]])

    phi1 = sopt.newton(optFunForEndPointR, 0, args=(Mrot, R, q0, k, sigma, 1))
    cp = np.cos(phi1)
    sp = np.sin(phi1)
    v = R*np.array([[cp - 1], [sp]])
    z1 = np.dot(Mrot, v)[0][0]

    phi2 = sopt.newton(optFunForEndPointR, 0, args=(Mrot, R, q0, k, sigma, -1))
    cp = np.cos(phi2)
    sp = np.sin(phi2)
    v = R*np.array([[cp - 1], [sp]])
    z2 = np.dot(Mrot, v)[0][0]

    return (z1,z2)

def optFunForFlat(a, Mrot, q0, k, sigma, side):
    '''
    A function to return the distance between the point on
    the spherical surface (flat?) at an angle phi and the beam width
    at the same z.

    Parameters
    ----------
    a : float
        a
    Mrot: array
        Rotational transformation.
    q0 : complex
        Beam parameter.
    k : float
        k
    sigma : float
        Beam width
    side
        side

    Returns
    -------
    float
        Distance between the point on the spherical surface at an
        angle phi and the beam width at the same z.
    '''
    v = np.array([[0],[a]])

    b = np.dot(Mrot, v)
    z = b[0]
    w1 = b[1]*side
    w2 = sigma*np.sqrt(-2.0/(k*np.imag(1/(q0+z))))
    return (w1-w2)[0]

def optimCrossPointFlat(theta, q0, k, sigma):
    '''Caltulate optimal cross point.

    Paramters
    ---------
    theta : float
        theta
    R : float
        R
    q0 : complex
        Beam parameter.
    k : float
        k
    sigma : float
        Beam width.

    Returns
    -------
    (float, float)
        Optimal end point.
    '''
    ct = np.cos(theta)
    st = np.sin(theta)
    Mrot = np.array([[ct, -st], [st, ct]])

    a1 = sopt.newton(optFunForFlat, 0, args=(Mrot, q0, k, sigma, 1))
    v = np.array([[0],[a1]])
    z1 = np.dot(Mrot, v)[0][0]

    a2 = sopt.newton(optFunForFlat, 0, args=(Mrot, q0, k, sigma, -1))
    v = np.array([[0],[a2]])
    z2 = np.dot(Mrot, v)[0][0]

    return (z1,z2)

#}}}



'''
Renderer module for gtrace.draw
'''

#{{{ Import modules

import gtrace.draw.dxf as dxf
import gtrace.draw as draw
from gtrace.unit import *
import numpy as np

#}}}

#{{{ Author and License Infomation

#Copyright (c) 2011-2021, Yoichi Aso
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

__author__ = "Yoichi Aso"
__copyright__ = "Copyright 2011-2021, Yoichi Aso"
__credits__ = ["Yoichi Aso"]
__license__ = "BSD"
__version__ = "0.2.4"
__maintainer__ = "Yoichi Aso"
__email__ = "yoichi.aso@nao.ac.jp"
__status__ = "Beta"

#}}}

#{{{ DXF renderer

class UnknownShapeError(BaseException):
    def __initi__(self, message):
        self.message = message

def renderDXF(canvas, filename):
    '''
    Render a canvas into a DXF file
    '''

    if canvas.unit == 'm':
        scale_factor = 1000.0
    elif canvas.unit == 'km':
        scale_factor = 1e6
    else:
        scale_factor = 1.0
        
    d = dxf.DXF(filename)

    for ly in list(canvas.layers.values()):
        d.add_layer(ly.name, color=dxf.color_encode(ly.color))
        for s in ly.shapes:
            if isinstance(s, draw.Line):
                d.add_entity(dxf.Line(np.array(s.start)*scale_factor, np.array(s.stop)*scale_factor), ly.name)
            elif isinstance(s, draw.PolyLine):
                x = np.array(s.x)*scale_factor
                y = np.array(s.y)*scale_factor
                d.add_entity(dxf.LwPolyLine(x, y), ly.name)
            elif isinstance(s, draw.Rectangle):
                d.add_entity(dxf.Rectangle(np.array(s.point)*scale_factor,
                                           s.width*scale_factor, s.height*scale_factor), ly.name)
            elif isinstance(s, draw.Circle):
                d.add_entity(dxf.Circle(np.array(s.center)*scale_factor, s.radius*scale_factor), ly.name)
            elif isinstance(s, draw.Arc):
                d.add_entity(dxf.Arc(np.array(s.center)*scale_factor, s.radius*scale_factor,
                                     rad2deg(s.startangle), rad2deg(s.stopangle)), ly.name)
            elif isinstance(s, draw.Text):
                d.add_entity(dxf.Text(s.text, np.array(s.point)*scale_factor, s.height*scale_factor, rad2deg(s.rotation)), ly.name)
            else:
                raise UnknownShapeError('Shape not supported')

    d.save_to_file()
                
#}}}



#{{{ Imports

import numpy as np
pi = np.pi

#}}}

#{{{ Length units
km = 1e3
m = 1.0
cm = 0.01
mm = 1e-3
um = 1e-6
nm = 1e-9
#}}}

#{{{ Power units

kW = 1e3
W = 1.0
mW = 1e-3
uW = 1e-6

#}}}

#{{{ Frequencies

THz = 1e12
GHz = 1e9
MHz = 1e6
kHz = 1e3
Hz = 1.0
mHz = 1e-3
uHz = 1e-6

#}}}

#{{{ Other

ppm = 1e-6

#}}}

#{{{ Conversion Functions

def rad2deg(rad):
    return rad * 180.0 / pi

def deg2rad(deg):
    return deg * pi /180.0

#}}}



"""
cavity.py - A Cavity class and related functions for representing a Fabry-Perot cavity
"""

#{{{ Import modules
import numpy as np
pi = np.pi
c = 299792458.0
sqrt = np.lib.scimath.sqrt
import gtrace.optics.gaussian as gauss
from traits.api import CFloat, HasTraits
#}}}

#{{{ Class Cavity

class Cavity(HasTraits):
    """
    A class to represent a Fabry-Perot cavity.

    Attributes
    ----------
    r1 : float
        Input mirror reflectivity (amplitude)
    r2 : float
        End mirror reflectivity (amplitude)
    rp1 : float
        Input mirror reflectivity (power)
    rp2 : float
        End mirror reflectivity (power)
    L : float
        Length
    R1 : float
        ROC of the input mirror (positive when concave to incident light, i.e.
        convex seen from inside the cavity)
    R2 : float
        ROC of the end mirror (positive when concave to incident light, i.e.
        concave seen from inside the cavity)
    wl : float
        Wavelength
    """
    r1 = CFloat(0.9)
    r2 = CFloat(0.99)
    rp1 = CFloat()
    rp2 = CFloat()
    tp1 = CFloat()
    tp2 = CFloat()
    t1 = CFloat()
    t2 = CFloat()

    L = CFloat(1.0)
    R1 = CFloat(-1.5)
    R2 = CFloat(1.5)
    g1 = CFloat()
    g2 = CFloat()
    wl=CFloat()

    def __init__(self, r1=0.9, r2=0.99, L=1.0, R1=-1.5, R2=1.5, wl=1064e-9, power=False):
        """
        Parameters
        ----------
        r1 : float, optional
            Input mirror reflectivity.
            Defaults 0.9.
        r2 : float, optional
            End mirror reflectivity.
            Defaults 0.99.
        L : float, optional.
            Length.
            Defaults 1.0.
        R1 : float, optional
            ROC of the input mirror (positive when concave to incident light, i.e.
            convex seen from inside the cavity)
            Defaults -1.5.
        R2 : float, optional
            ROC of the end mirror (positive when concave to incident light, i.e.
            concave seen from inside the cavity).
            Default 1.5.
        wl : float, optional
            Wavelength.
            Defaults 1064e-9.
        power : boolean, optional
            If True, r1 and r2 are treated as power reflectivities.
            Otherwise, r1 and r2 are regarded as amplitude reflectivities.
            Defaults False.
        """
        if power:
            self.r1 = np.sqrt(r1)
            self.r2 = np.sqrt(r2)
        else:
            self.r1 = r1
            self.r2 = r2

        self.L = L
        self.R1 = R1
        self.R2 = R2
        self.wl = wl

    #{{{ Trait change handlers
    def _R1_changed(self, old, new):
        self.g1 = 1 + self.L/new

    def _R2_changed(self, old, new):
        self.g2 = 1 - self.L/new

    def _L_changed(self, old, new):
        self.g1 = 1 + new/self.R1
        self.g2 = 1 - new/self.R2

    def _r1_changed(self, old, r1):
        self.rp1 = r1**2
        self.tp1 = 1 - self.rp1
        self.t1 = np.sqrt(self.tp1)

    def _r2_changed(self, old, r2):
        self.rp2 = r2**2
        self.tp2 = 1 - self.rp2
        self.t2 = np.sqrt(self.tp2)

    #}}}

    def finesse(self):
        '''
        Returns the finesse of the cavity.
        '''
        return finesse(self.r1, self.r2)

    def storageTime(self):
        '''
        Storage time
        '''

        return 2*self.L*self.finesse()/(2*pi*c)

    def pole(self):
        '''
        Cavity pole frequency [Hz]
        '''

        return 1/(2*pi*self.storageTime())

    def Nbounce(self):
        '''
        Bounce number
        '''

        return 2*self.finesse()/pi

    def powerGain(self):
        '''
        Ratio of the intra-cavity power to the input power.
        '''
        return (self.t1/(1-self.r1*self.r2))**2

    def FSR(self):
        '''
        Returns the free spectral range of the cavity.
        '''
        return c/(2*self.L)

    def modeSpacing(self):
        '''
        Return the transverse mode spacing of the cavity (commonly called gamma).
        It is a fractional number defined by gamma = (mode spacing frequency)/FSR.
        '''
        return gauss.modeSpacing(self.g1, self.g2)

    def waist(self, size=False):
        """
        Return the q-parameter or the radius of the beam at the cavity waist.

        Parameters
        ----------
        size : boolean, optional
            if set to true, the first element of the returned tuple will be the waist size, rather than the q-parameter.

        Returns
        -------
        (q0, d) : (complex, float)
            This function returns a tuple with two elements.
            The first element is the q-parameter of the cavity mode at
            the cavity waist. If size=True is given, it becomes the waist
            size (1/e^2 radius).
            The second element is the distance of the cavity waist from
            the input mirror.
        """
        # q0 = 1j*np.sqrt(self.L)*np.sqrt(-(self.L+self.R1)*(self.L-self.R2)\
        #                                 *(self.L+self.R1-self.R2))\
        #                                 /(2*self.L+self.R1-self.R2)

        q0 = 1j*sqrt(-self.L*(self.L+self.R1)**2*(self.L-self.R2)*\
                         (self.L+self.R1-self.R2)/((-2*self.L-self.R1+self.R2)**2))\
                         /sqrt(self.L+self.R1)

        d = self.L*(self.L-self.R2)/(2*self.L+self.R1-self.R2)

        if size:
            return (gauss.q2w(q0, wl=self.wl), d)
        else:
            return (q0, d)

    def spotSize(self):
        """
        Returns the beam spot sizes on the input and end mirrors
        as a tuple (w1,w2).
        """
        (q0,d) = self.waist()
        w1=gauss.q2w(q0-d, wl=self.wl)
        w2=gauss.q2w(q0+self.L-d, wl=self.wl)
        return (w1, w2)

    def trans(self, f=0, d=0):
        """
        Returns the amplitude transmissivity of the cavity.
        It assumes the cavity was locked to the incident light first. Then computes the
        amplitude transmissivity for the light with a frequency shift f from the original light
        with the cavity length changed by d from the initial state.

        Parameters
        ----------
        f : float, optional
            Frequency shift of the light in Hz.
            Defaults 0.
        d : float, optional.
            Cavity length detuning in m.
            Defaults 0.

        Returns
        -------
        complex
            The amplitude transmissivity of the cavity (a complex number).

        """
        #One way phase change
        phi=2*pi*(self.L*f/c + d/self.wl + d*f/c)

        return self.t1 * self.t2 * np.exp(-1j*phi)/\
                   (1-self.r1*self.r2*np.exp(-2j*phi))

    def refl(self, f=0, d=0):
        """
        Returns the amplitude reflectivity of the cavity.
        It assumes the cavity was locked to the incident light first. Then computes the
        amplitude reflectivity for the light with a frequency shift f from the original light
        with the cavity length changed by d from the initial state.

        Parameters
        ----------
        f : float, optional
            Frequency shift of the light in Hz.
            Defaults 0.
        d : float, optional.
            Cavity length detuning in m.
            Defaults 0.

        Returns
        -------
        complex
            The amplitude reflectivity of the cavity (a complex number).
        """

        #One way phase change
        phi=2*pi*(self.L*f/c + d/self.wl + d*f/c)

        return -self.r1+self.t1**2 * self.r2 *np.exp(-2j*phi)/\
                   (1-self.r1*self.r2*np.exp(-2j*phi))

    def intra(self, f=0, d=0):
        """
        Returns the intra cavity field amplitude.
        It assumes the cavity was locked to the incident light first. Then computes the
        intra-cavity field amplitude for the light with a frequency shift f from the original light
        with the cavity length changed by d from the initial state.

        Parameters
        ----------
        f : float, optional
            Frequency shift of the light in Hz.
            Defaults 0.
        d : float, optional.
            Cavity length detuning in m.
            Defaults 0.

        Returns
        -------
        complex
            The intra-cavity field amplitude at the input mirror surface (a complex number).

        """
        #One way phase change
        phi=2*pi*((self.L+d)*f/c + d/self.wl)

        return self.t1/(1-self.r1*self.r2*np.exp(-2j*phi))

#}}}

#{{{ Often used parameters

def finesse(r1, r2, power=False):
    '''
    Returns the finesse of a cavity

    Parameters
    ----------
    r1 : float
        Reflectivity of the first mirror.
    r2 : float
        Reflectivity of the second mirror.
    power : boolean, optional
        If True, r1 and r2 are treated as power reflectivities.
        Otherwise, r1 and r2 are regarded as amplitude reflectivities.
        Defaults False.
    '''
    if power:
        r1 = np.sqrt(r1)
        r2 = np.sqrt(r2)

    return pi*np.sqrt(r1*r2)/(1-r1*r2)

#}}}



'''
Generate example outputs
'''

#{{{ Import Modules
#First import modules

import gtrace.beam as beam
import gtrace.optcomp as opt
from gtrace.draw.tools import drawAllBeams, drawAllOptics, transAll, rotateAll
import gtrace.draw as draw
import gtrace.draw.renderer as renderer
from gtrace.unit import *
import gtrace.optics.gaussian as gauss
from gtrace.nonsequential import non_seq_trace

#Numpy is almost mandatory to use gtrace
import numpy as np
pi = np.pi

#}}}

#{{{ GaussianBeam object

#q-parameter of the beam
q0 = gauss.Rw2q(ROC=np.inf, w=0.3*mm)

#Create a GaussianBeam object.
b1 = beam.GaussianBeam(q0=q0, wl=1064*nm, length=30*cm, P=1.0)

#Set the direction angle of the beam to 10deg from the global x-axis.
b1.dirAngle = deg2rad(10)

#Set the position of the origin of the beam
b1.pos = (0.0, 0.0)

#}}}

#{{{ Mirror object

M1 = opt.Mirror(HRcenter=[50*cm, 10*cm], normAngleHR=pi,
                 diameter=25*cm, thickness=10*cm,
                 wedgeAngle=deg2rad(0.25), inv_ROC_HR=1./(120*cm),
                 inv_ROC_AR=0,
                 Refl_HR=0.9, Trans_HR=1-0.9,
                 Refl_AR=500*ppm, Trans_AR=1-500*ppm,
                 n=1.45, name='M1')

M2 = opt.Mirror(HRcenter=[0*cm, 18*cm], normAngleHR=deg2rad(5.0),
                 diameter=15*cm, thickness=5*cm,
                 wedgeAngle=deg2rad(0.25), inv_ROC_HR=-1./(350*cm),
                 inv_ROC_AR=0,
                 Refl_HR=0.9, Trans_HR=1-0.9,
                 Refl_AR=500*ppm, Trans_AR=1-500*ppm,
                 n=1.45, name='M2')

M3 = opt.Mirror(HRcenter=[30*cm, 30*cm], normAngleHR=deg2rad(21.3),
                 diameter=15*cm, thickness=5*cm,
                 wedgeAngle=deg2rad(1), inv_ROC_HR=1./(350*cm),
                 inv_ROC_AR=0,
                 Refl_HR=0.9, Trans_HR=1-0.9,
                 Refl_AR=500*ppm, Trans_AR=1-500*ppm,
                 n=1.45, name='M3')


#}}}

#{{{ Sequential

#{{{ Hit the mirror with the beam

#beams is a dictionary of the beams created by
#the injection of the beam into the mirror
beams1 = M1.hitFromHR(b1, order=2)
b2 = beams1.pop('r1')
b2.length=20*cm
beams2 = M2.hitFromHR(b2, order=2)
b3 = beams2.pop('r1')
b3.length=20*cm
beams3 = M3.hitFromAR(b3, order=2)

#}}}

#{{{ Draw

#Create a canvas object
cnv = draw.Canvas()

#Draw all beams

drawAllBeams(cnv, beams1.values(), drawWidth=True, sigma=3.0, drawPower=False,
                 drawROC=False, drawGouy=False, drawOptDist=False, layer='main_beam1',
                    fontSize=0.01)

drawAllBeams(cnv, beams2.values(), drawWidth=True, sigma=3.0, drawPower=False,
                 drawROC=False, drawGouy=False, drawOptDist=False, layer='main_beam2',
                    fontSize=0.01)

drawAllBeams(cnv, beams3.values(), drawWidth=True, sigma=3.0, drawPower=False,
                 drawROC=False, drawGouy=False, drawOptDist=False, layer='main_beam3',
                    fontSize=0.01)

drawAllBeams(cnv, [b1,b2,b3], drawWidth=True, sigma=3.0, drawPower=False,
                 drawROC=False, drawGouy=False, drawOptDist=False, layer='source_beam',
                    fontSize=0.01)

#Draw the mirror
drawAllOptics(cnv, [M1,M2,M3])

#Save the result as a DXF file
renderer.renderDXF(cnv, 'SeqTrace.dxf')

#}}}

#}}}

#{{{ Non-Sequential

#{{{ Hit the mirrors

beams = non_seq_trace([M1,M2,M3], b1, order=30, power_threshold=1e-6)

#}}}

#{{{ Draw

#Create a canvas object
cnv = draw.Canvas()

#Draw all beams

drawAllBeams(cnv, beams, drawWidth=True, sigma=3.0, drawPower=False,
                 drawROC=False, drawGouy=False, drawOptDist=False, layer='main_beam',
                    fontSize=0.01)

drawAllBeams(cnv, [b1], drawWidth=True, sigma=3.0, drawPower=False,
                 drawROC=False, drawGouy=False, drawOptDist=False, layer='source_beam',
                    fontSize=0.01)

#Draw the mirror
drawAllOptics(cnv, [M1,M2,M3])

#Save the result as a DXF file
renderer.renderDXF(cnv, 'NonSeq.dxf')

#}}}

#}}}

#{{{ Misc

#{{{ A Gaussian Beam

#q-parameter of the beam
q0 = gauss.Rw2q(ROC=np.inf, w=0.03*mm)

#Create a GaussianBeam object.
b1 = beam.GaussianBeam(q0=q0, wl=1064*nm, length=6*cm, P=1.0)

b1.propagate(-3*cm)

#Set the direction angle of the beam to 10deg from the global x-axis.
b1.dirAngle = deg2rad(0)

#Set the position of the origin of the beam
b1.pos = (0.0, 0.0)

#Create a canvas object
cnv = draw.Canvas()

#Draw all beams

drawAllBeams(cnv, [b1], drawWidth=True, sigma=3.0, drawPower=False,
                 drawROC=False, drawGouy=False, drawOptDist=False, layer='source_beam',
                    fontSize=0.01)


#Save the result as a DXF file
renderer.renderDXF(cnv, 'GaussBeam.dxf')

#}}}

#}}}





#{{{ Import Modules
#First import modules

import gtrace.beam as beam
import gtrace.optcomp as opt
from gtrace.draw.tools import drawAllBeams, drawAllOptics, transAll, rotateAll
import gtrace.draw as draw
import gtrace.draw.renderer as renderer
from gtrace.unit import *
import gtrace.optics.gaussian as gauss

#Numpy is almost mandatory to use gtrace
import numpy as np
pi = np.pi

#}}}

#{{{ GaussianBeam object
#First, we create a beam object.

#Prepare a q-parameter
#Rw2q() converts ROC and the radius of a beam
#into a q-parameter. np.inf is infinity.
q0 = gauss.Rw2q(np.inf, 0.1*mm)

#Create a GaussianBeam object.
b = beam.GaussianBeam(q0=q0, wl=1064*nm)

#Set the direction angle of the beam to 10deg from the global x-axis.
b.dirAngle = deg2rad(10)

#Set the position of the origin of the beam
b.pos = (0.0, 0.0)

#}}}

#{{{ Mirror object
#Put a mirror 10cm away from the beam origin
M1 = opt.CyMirror(HRcenter=[0.1, 0], normAngleHR=pi,
                 diameter=25*cm, thickness=10*cm,
                 wedgeAngle=deg2rad(2.5), inv_ROC_HR=1./1,
                 inv_ROC_AR=-1./2,
                 Refl_HR=0.9, Trans_HR=1-0.9,
                 Refl_AR=500*ppm, Trans_AR=1-500*ppm,
                 n=1.45, name='Mirror', curve_direction='h')


M1.rotate(deg2rad(-45))

#}}}

#{{{ Hit the mirror with the beam

#beams is a dictionary of the beams created by
#the injection of the beam into the mirror
beams = M1.hitFromHR(b, order=3)

#}}}

#{{{ Draw

#Create a canvas object
cnv = draw.Canvas()

#Add layers (optional)
cnv.add_layer("main_beam", color=(0,0,0))

#Draw all beams
beamList = beams.values()
drawAllBeams(cnv, beamList, drawWidth=True, sigma=3.0, drawPower=False,
                 drawROC=False, drawGouy=False, drawOptDist=False, layer='main_beam',
                    fontSize=0.01)

#You can also draw each beam manually. For example,
#b1 = beams['s1']
#b1.draw(cnv)
#This will draw the beam b1 into the canvas object cnv.

#Draw the mirror
M1.draw(cnv, drawName=True)

#Save the result as a DXF file
renderer.renderDXF(cnv, 'test.dxf')

#}}}

#{{{ Simple
q0 = gauss.Rw2q(np.inf, 0.1*mm)
b = beam.GaussianBeam(q0=q0, wl=1064*nm, pos=[0.0,0.0], dirAngle=0.0)

M1 = opt.Mirror(HRcenter=[2, 0], normAngleHR=deg2rad(170),
                 diameter=25*cm, thickness=10*cm,
                 wedgeAngle=deg2rad(0.25), inv_ROC_HR=1./2.3,
                 Refl_HR=0.9, Trans_HR=1-0.9,
                 n=1.45, name='Mirror')

beams = M1.hitFromHR(b, order=2)

cnv = draw.Canvas()
cnv.add_layer("main_beam", color=(0,0,0))
cnv.add_layer("main_beam_width", color=(0,0,1))
drawAllBeams(cnv, beams.values(), drawWidth=True, sigma=2.7,  layer='main_beam')
M1.draw(cnv, drawName=True)
renderer.renderDXF(cnv, 'Test.dxf')

#}}}



#{{{ Imports

import numpy as np
pi = np.pi

#}}}

#{{{ Author and License Infomation

#Copyright (c) 2011, Yoichi Aso
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

__author__ = "Yoichi Aso"
__copyright__ = "Copyright 2011, Yoichi Aso"
__credits__ = ["Yoichi Aso"]
__license__ = "BSD"
__version__ = "0.2.4"
__maintainer__ = "Yoichi Aso"
__email__ = "yoichi.aso@nao.ac.jp"
__status__ = "Beta"

#}}}

#{{{ Length units
km = 1e3
m = 1.0
cm = 0.01
mm = 1e-3
um = 1e-6
nm = 1e-9
#}}}

#{{{ Power units

kW = 1e3
W = 1.0
mW = 1e-3
uW = 1e-6

#}}}

#{{{ Frequencies

THz = 1e12
GHz = 1e9
MHz = 1e6
kHz = 1e3
Hz = 1.0
mHz = 1e-3
uHz = 1e-6

#}}}

#{{{ Other

ppm = 1e-6
percent = 1e-2

#}}}

#{{{ Conversion Functions

def rad2deg(rad):
    return rad * 180.0 / pi

def deg2rad(deg):
    return deg * pi /180.0

#}}}



'''
Drawing classes for gtrace
'''

#{{{ Import modules
import numpy as np
pi = np.pi
#}}}

#{{{ Author and License Infomation

#Copyright (c) 2011-2021, Yoichi Aso
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

__author__ = "Yoichi Aso"
__copyright__ = "Copyright 2011-2021, Yoichi Aso"
__credits__ = ["Yoichi Aso"]
__license__ = "BSD"
__version__ = "0.2.4"
__maintainer__ = "Yoichi Aso"
__email__ = "yoichi.aso@nao.ac.jp"
__status__ = "Beta"

#}}}

#{{{ Canvas class

class Canvas(object):
    '''
    Canvas class
    '''

    def __init__(self, unit='m'):
        self.layers = {}
        self.unit = unit

    def add_layer(self, name, color=(0,0,0)):
        self.layers[name] = Layer(name, color=color)

    def add_shape(self, shape, layername):
        if not layername in self.layers:
            self.add_layer(layername)

        self.layers[layername].add_shape(shape)

#}}}

#{{{ Layer class

class Layer(object):
    '''
    Layer class
    '''

    def __init__(self, name, color=(0,0,0)):
        self.name = name
        self.color = color
        self.shapes = []

    def add_shape(self, shape):
        self.shapes.append(shape)
        
#}}}

#{{{ Shape class

class Shape(object):
    '''
    Shape class
    '''

    def __init__(self):
        pass
        
#}}}

#{{{ Line class

class Line(Shape):
    '''
    Line class
    '''

    def __init__(self, start, stop, thickness=0):
        super(Line, self).__init__()
        self.start = start
        self.stop = stop
        self.thickness = thickness
        
#}}}

#{{{ PolyLine

class NumberOfElementError(BaseException):
    def __initi__(self, message):
        self.message = message
        
class PolyLine(Shape):
    '''
    A light weight poly-line
    '''
    
    def __init__(self, x, y, thickness=0):
        '''
        = Arguments =
        x: x coordinates of the vertices
        y: y coordinates of the vertices        
        '''
        super(PolyLine, self).__init__()
        self.x = x
        self.y = y
        if len(x) != len(y):
            raise NumberOfElementError('The numbers of elements of x and y do not match.')
        self.numpoints = len(x)
        self.thickness = thickness


#}}}

#{{{ Rectangle

class Rectangle(Shape):
    '''
    A rectangle
    '''
    
    def __init__(self, point, width, height, thickness=0):
        '''
        = Arguments =
        point: lower left corner of the rectangle
        width:
        height:
        '''
        super(Rectangle, self).__init__()
        self.point = point
        self.width = width
        self.height = height
        self.thickness = thickness


#}}}

#{{{ Circle

class Circle(Shape):
    '''
    A circle
    '''

    def __init__(self, center, radius, thickness=0):
        super(Circle, self).__init__()
        self.center = center
        self.radius = radius
        self.thickness = thickness

#}}}

#{{{ Arc

class Arc(Shape):
    '''
    An arc

    Note that angles are stored in rad.
    '''

    def __init__(self, center, radius, startangle, stopangle, thickness=0, angle_in_rad=True):
        super(Arc, self).__init__()
        self.center = center
        self.radius = radius
        self.thickness = thickness
        if angle_in_rad:
            self.startangle = startangle
            self.stopangle = stopangle
        else:
            self.startangle = pi*startangle/180.0
            self.stopangle = pi*stopangle/180.0
            

#}}}

#{{{ Text

class Text(Shape):
    '''
    Text

    Note that angles are stored in rad.    
    '''

    def __init__(self, text, point, height=1.0, rotation=0.0, angle_in_rad=True):
        super(Text, self).__init__()
        self.text = text
        self.point = point
        self.height = height
        if angle_in_rad:        
            self.rotation = rotation
        else:
            self.rotation = pi*rotation/180.0

#}}}



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gtrace",
    version="0.2.4",
    author="Yoichi Aso",
    author_email="yoichi.aso@nao.ac.jp",
    description="2D ray tracing package for Gaussian beams",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/asoy01/gtrace",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires = ['numpy>=1.5.0', 'scipy>=0.1.0','traits>=4.0.0']
)





#{{{ ==== Import modules ====

#Add the local "modules" directory to the python's module search path.
#This trick only works if you run this script from the root directory of OptLayout.
import sys
import os
sys.path.append(os.path.abspath(os.curdir))

from matplotlib.font_manager import FontProperties, FontManager
import gtrace.beam as beam
import gtrace.optcomp as opt
from gtrace.nonsequential import non_seq_trace
from gtrace.draw.tools import drawAllBeams, drawAllOptics, transAll, rotateAll, drawOptSys
import gtrace.draw as draw
import gtrace.draw.renderer as renderer
from gtrace.unit import *
import gtrace.optics.gaussian as gauss
from gtrace.optics.cavity import Cavity
import gtrace.optics.geometric as geom
import numpy as np
pi = np.pi
import scipy as sp
import scipy.optimize as sopt
import copy
from scipy.constants import c
from multiprocessing import Process, Pipe, cpu_count
import time
import sharedmem as shm
from paralleltools import distribute1DArray, parallelize
import pickle

#}}}

#{{{ Parameters

MC_Dia = 10.0*cm
MC_Thick = 3.0*cm
MCe_ROC = 40.0
MMT1_ROC = -24.6
MMT2_ROC = 24.3

MCi_Refl = 0.9937
MCo_Refl = 0.9937
MCe_Refl = 0.9999
AR_Refl = 0.1/100

pos_MCi = np.array([-27.759072, 0.136953797]) 
pos_MCo = np.array([-27.259072, 0.136953797]) 
pos_MCe = np.array([-27.509048, 26.53565317]) 
pos_MMT1 = np.array([-21.394048, -0.060390856]) 
pos_MMT2 = np.array([-24.479048, 0.290123756]) 
LMMT = np.linalg.norm(pos_MMT1 - pos_MMT2)
L_MC_MMT1 = 5.8357910693793

Lmc = (np.linalg.norm(pos_MCi - pos_MCo) + np.linalg.norm(pos_MCi - pos_MCe) + np.linalg.norm(pos_MCo - pos_MCe))/2

MC = Cavity(r1=0.9, r2=0.9, L=Lmc, R1=-1e8, R2=MCe_ROC, power=True)

MCq0 = MC.waist()[0]

inputBeamDict = {}



nsilica = 1.44967
nsilica_green = 1.46071
nsaph = 1.754
nsaph_green = 1.7717



#}}}

#{{{ Define mirrors


MCi = opt.Mirror(HRcenter=[0,0], normAngleHR=0.0,
                 diameter=MC_Dia, thickness=MC_Thick,
                 wedgeAngle=-deg2rad(2.5), inv_ROC_HR=0.0,
                 Refl_HR=MCi_Refl, Trans_HR=1-MCi_Refl,
                 Refl_AR=AR_Refl, Trans_AR=1-AR_Refl,
                 n=nsilica, name='MCi')

MCo = opt.Mirror(HRcenter=[0,0], normAngleHR=0.0,
                 diameter=MC_Dia, thickness=MC_Thick,
                 wedgeAngle=deg2rad(2.5), inv_ROC_HR=0.0,
                 Refl_HR=MCo_Refl, Trans_HR=1-MCo_Refl,
                 Refl_AR=AR_Refl, Trans_AR=1-AR_Refl,
                 n=nsilica, name='MCo')

MCe = opt.Mirror(HRcenter=[0,0], normAngleHR=0.0,
                 diameter=MC_Dia, thickness=MC_Thick,
                 wedgeAngle=deg2rad(2.5), inv_ROC_HR=1.0/MCe_ROC,
                 Refl_HR=MCe_Refl, Trans_HR=1-MCe_Refl,
                 Refl_AR=AR_Refl, Trans_AR=1-AR_Refl,
                 n=nsilica, name='MCe')

MMT1 = opt.Mirror(HRcenter=[0,0], normAngleHR=0.0,
                 diameter=MC_Dia, thickness=MC_Thick,
                 wedgeAngle=deg2rad(2.5), inv_ROC_HR=1.0/MMT1_ROC,
                 Refl_HR=0.999, Trans_HR=1-0.999,
                 Refl_AR=AR_Refl, Trans_AR=1-AR_Refl,
                 n=nsilica, name='MMT1')

MMT2 = opt.Mirror(HRcenter=[0,0], normAngleHR=0.0,
                 diameter=MC_Dia, thickness=MC_Thick,
                 wedgeAngle=deg2rad(2.5), inv_ROC_HR=1.0/MMT2_ROC,
                 Refl_HR=0.999, Trans_HR=1-0.999,
                 Refl_AR=AR_Refl, Trans_AR=1-AR_Refl,
                 n=nsilica, name='MMT2')


inputOptics = [MCi, MCo, MCe, MMT1, MMT2]

#}}}

#{{{ Put MC mirrors

MCi.HRcenter = pos_MCi
MCo.HRcenter = pos_MCo
MCe.HRcenter = pos_MCe

v1 = pos_MCi - pos_MCo
v2 = pos_MCe - pos_MCo
v1 = v1/np.linalg.norm(v1)
v2 = v2/np.linalg.norm(v2)
MCo.normVectHR = (v1+v2)/2

v2 = pos_MCe - pos_MCi
v2 = v2/np.linalg.norm(v2)
MCi.normVectHR = (-v1+v2)/2

MCe.normVectHR = np.array([0.0, -1.0])

#}}}

#{{{ Propagate MC beams

bmc = beam.GaussianBeam(q0 = MCq0, pos=(pos_MCi + pos_MCo)/2, dirAngle=0.0)

beams = MCo.hitFromHR(bmc, order=1)

b1=beams['input'].copy()
b1.propagate(b1.length)
b2 = beams['t1'].copy()

L=np.real(b2.qy - b1.qy)*nsilica

theta1=0.780663677706396
theta2=0.5068847403834057
theta3=0.46325150908354695
theta4=0.70469103773435282

c1=np.cos(theta1)
c2=np.cos(theta2)
c3=np.cos(theta3)
c4=np.cos(theta4)

a=c4*c2/(c3*c1)
b=c4*c1*L/(c3*c2*nsilica)
c=0
d=c1*c3/(c2*c4)
q=b1.qx

(a*q+b)/(c*q+d)

#}}}
